from typing import Any, Callable, List, Optional, Union

from aiotfm.enums import GameMode
from aiotfm.errors import AiotfmException
from aiotfm.packet import Packet
from aiotfm.player import Player


class Room:
	"""Represents the room that the bot currently is in.

	Attributes
	----------
	name: `str`
		The room's name. (i.e: en-1, *bad girls and so on)
	official: `bool`
		Whether the room is an official room or not. If official, it's name will be displayed in yellow.
	players: `list[:class:`aiotfm.Player`]`
		The list containing all the players of the room.
	"""
	def __init__(self, name: str, official: bool = False):
		self.name: str = name
		self.official: bool = official
		self.players: dict = {}

	def __repr__(self):
		return "<Room name={} official={}>".format(self.name, self.official)

	@property
	def community(self) -> str:
		"""Returns the room's community."""
		if self.name.startswith('*'):
			return 'xx'
		return self.name.split('-', 1)[0]

	@property
	def is_tribe(self) -> bool:
		"""Returns true if it's a tribe house."""
		return self.name.startswith('*\x03')

	@property
	def display_name(self) -> str:
		r"""Return the display name of the room.
		It removes the \x03 char from the tribe house and the community from the public rooms."""
		if self.is_tribe:
			return self.name.replace('\x03', '')
		if self.name.startswith('*'):
			return self.name
		return self.name.split('-', 1)[1]

	def get_players(self, predicate: Callable, max_: Optional[int] = None) -> List[Player]:
		"""Filters players from the room.

		:param predicate: A function that returns a boolean-like result to filter through
			the players.
		:param max_: Optional[:class:`int`] The maximum amount of players to return.
		:return: `Iterable` The filtered players."""
		return [p for p in self.players.values() if predicate(p)][:max_]

	def get_player(self, default: Optional[Any] = None, **kwargs) -> Union[Player, Any]:
		"""Gets one player in the room with an identifier.

		:param kwargs: Which identifier to use. Can be either name, username, id or pid.
		:return: :class:`aiotfm.Player` The player or None"""
		length = len(kwargs.keys())

		if length == 0:
			raise AiotfmException('You did not provide any identifier.')
		if length > 1:
			raise AiotfmException('You cannot filter one player with more than one identifier.')

		identifier, value = next(iter(kwargs.items()))

		if identifier in ('name', 'username'):
			def filter_(p):
				return p == value
		elif identifier == 'id':
			def filter_(p):
				return p.id == int(value)
		elif identifier == 'pid':
			return self.players.get(int(value), default)
		else:
			raise AiotfmException('Invalid filter.')

		for player in self.players.values():
			if filter_(player):
				return player
		return default


class RoomEntry:
	__slots__ = (
		'name', 'language', 'country', 'player_count', 'limit',
		'is_funcorp', 'is_pinned', 'command', 'args', 'is_modified',
		'shaman_skills', 'consumables', 'adventure', 'collision',
		'aie', 'map_duration', 'mice_mass', 'map_rotation'
	)

	def __init__(
		self, name: str, language: str, country: str, player_count: int,
		limit: int = 0, is_funcorp: bool = False, is_pinned: bool = False,
		command: str = '', args: str = '', is_modified: bool = False, shaman_skills: bool = True,
		consumables: bool = True, adventure: bool = True, collision: bool = False, aie: bool = False,
		map_duration: int = 100, mice_mass: int = 100, map_rotation: list = []
	):
		self.name: str = name
		self.language: str = language
		self.country: str = country
		self.player_count: int = player_count
		self.limit: int = limit
		self.is_funcorp: bool = is_funcorp
		self.is_pinned: bool = is_pinned
		self.command: str = command
		self.args: str = args
		self.is_modified: bool = is_modified
		self.shaman_skills: bool = shaman_skills
		self.consumables: bool = consumables
		self.adventure: bool = adventure
		self.collision: bool = collision
		self.aie: bool = aie
		self.map_duration: int = map_duration
		self.mice_mass: int = mice_mass
		self.map_rotation: list = map_rotation

	def __repr__(self):
		return '<{} {}>'.format(
			self.__class__.__name__,
			' '.join('{}={!r}'.format(key, getattr(self, key)) for key in self.__slots__)
		)


class DropdownRoomEntry(RoomEntry):
	__slots__ = ('entries',)

	def __init__(self, entries, *args, **kwargs):
		super().__init__(*args, **kwargs, is_pinned=True)
		self.entries = entries


class RoomList:
	"""Represents the list of rooms in the server.

	Attributes
	----------
	gamemode: :class:`aiotfm.enums.GameMode`
		The list's gamemode.
	rooms: List[`RoomEntry`]
		The list of normal rooms.
	pinned_rooms: List[`RoomEntry`]
		The list of pinned(/module) rooms.
	gamemodes: List[:class:`aiotfm.enums.GameMode`]
		The list of gamemodes available.
	"""
	def __init__(
		self, gamemode: GameMode, rooms: List[RoomEntry],
		pinned_rooms: List[RoomEntry], gamemodes: List[GameMode]
	):
		self.gamemode: GameMode = gamemode
		self.rooms: List[RoomEntry] = rooms
		self.pinned_rooms: List[RoomEntry] = pinned_rooms
		self.gamemodes: List[GameMode] = gamemodes

	@classmethod
	def from_packet(cls, packet: Packet):
		gamemodes = [GameMode(packet.read8()) for _ in range(packet.read8())]
		gamemode = GameMode(packet.read8())
		rooms: List[RoomEntry] = []
		pinned: List[RoomEntry] = []

		while packet.pos < len(packet.buffer):
			is_pinned = packet.readBool()
			language = packet.readUTF()
			country = packet.readUTF()
			name = packet.readUTF()

			if is_pinned:
				player_count = packet.readUTF()
				command = packet.readUTF()
				args = packet.readUTF()

				if player_count.isdigit():
					player_count = int(player_count)

				if command == 'lm':
					entries: List[RoomEntry] = []
					room = DropdownRoomEntry(entries, name, language, country, player_count)

					for mode in args.split('&~'):
						if ',' not in mode:
							continue

						name, count = mode.split(',')
						entries.append(RoomEntry(
							name, room.language, room.country, int(count),
							command='mjj', args='m ' + name
						))

					pinned.append(room)
				else:
					pinned.append(RoomEntry(
						name, language, country, player_count,
						command=command, args=args, is_pinned=True
					))
			else:
				player_count = packet.read16()
				limit = packet.read8()
				is_funcorp = packet.readBool()
				is_modified = packet.readBool()

				kwargs = {
					"limit": limit,
					"is_funcorp": is_funcorp,
				}

				# Read the modified properties
				if is_modified:
					shaman_skills = not packet.readBool()
					consumables = not packet.readBool()
					adventure = not packet.readBool()
					collision = packet.readBool()
					aie = packet.readBool()
					map_duration = packet.read8()
					mice_mass = packet.read32()
					map_rotation = []

					for i in range(0, packet.read8()):
						map_rotation.append(packet.read8())

					# Append the room's specific properties
					kwargs.update({
						"is_modified": is_modified,
						"shaman_skills": shaman_skills,
						"consumables": consumables,
						"adventure": adventure,
						"collision": collision,
						"aie": aie,
						"map_duration": map_duration,
						"mice_mass": mice_mass,
						"map_rotation": map_rotation
					})

				rooms.append(RoomEntry(
					name, language, country, player_count,
					**kwargs
				))

		return cls(gamemode, rooms, pinned, gamemodes)



from datetime import datetime


class Date(datetime):
	"""Represents the date format of the game.

	Inherit from datetime.datetime class.
	"""
	@classmethod
	def fromtimestamp(cls, t, tz=None):
		"""Return a date from a timestamp
		:param t: :class:`int` the timestamp.
		:param tz: Optional[:class:`datetime.timezone`] a time zone to pass to the super method.
		:return: :class:`aiotfm.utils.Date` the date.
		"""
		return cls(*super().fromtimestamp(t * 60, tz).timetuple()[:6])

	def timestamp(self):
		"""Convert the date into the timestamp format used by the game.
		:return: :class:`int` the timestamp.
		"""
		return round(super().timestamp() / 60)



from aiotfm.packet import Packet


class Player:
	"""Represents a player in game.

	Attributes
	----------
	username: :class:`str`
		The player's username.
	uid: :class:`int`
		The player's id. -1 if unknown
	pid: :class:`int`
		The player's pid. -1 if unknown
	look: :class:`str`
		The player's look. '' if unknown
	gender: :class:`int`
		The player's gender.
	title: :class:`int`
		The player's title id. 0 if unknown
	title_stars: :class:`int`
		The player's title's stars.
	cheeses: :class:`int`
		Amount of cheese the player has.
	isDead: :class:`bool`
		True if the player is dead.
	isShaman: :class:`bool`
		True if the player is shaman.
	isVampire: :class:`bool`
		True if the player is vampire.
	score: :class:`int`
		The player's score.
	mouseColor: :class:`int`
		The color of the player's fur.
	nameColor: :class:`int`
		The color of the player's name.
	shamanColor: :class:`int`
		The color of the player's shaman's feather.
	facingRight: :class:`bool`
		True if the player is facing right.
	movingLeft: :class:`bool`
		True if the player is moving to the left.
	movingRight: :class:`bool`
		True if the player is moving to the right.
	x: :class:`int`
		The player's x position.
	y: :class:`int`
		The player's y position.
	vx: :class:`int`
		The player's horizontal speed.
	vy: :class:`int`
		The player's vertical speed.
	ducking: :class:`bool`
		True if the player is ducking (crouching).
	jumping: :class:`bool`
		True if the player is jumping.
	"""
	def __init__(self, username, uid=-1, pid=-1, **kwargs):
		self.gender = kwargs.get('gender', 0)
		self.look = kwargs.get('look', '')
		self.id = uid
		self.pid = pid
		self.title = kwargs.get('title', 0)
		self.title_stars = kwargs.get('title_stars', 0)
		self.username = username

		self.cheeses = kwargs.get('cheeses', 0)
		self.isDead = kwargs.get('isDead', False)
		self.isShaman = kwargs.get('isShaman', False)
		self.isVampire = kwargs.get('isVampire', False)
		self.score = kwargs.get('score', 0)

		self.mouseColor = kwargs.get('mouseColor', 0)
		self.nameColor = kwargs.get('nameColor', -1)
		self.shamanColor = kwargs.get('shamanColor', 0)

		self.facingRight = True
		self.movingLeft = False
		self.movingRight = False

		self.x = 0
		self.y = 0
		self.vx = 0
		self.vy = 0
		self.ducking = False
		self.jumping = False

	def __repr__(self):
		info = ' '.join([f'{field}={getattr(self, field)!r}' for field in ('id', 'pid') if getattr(self, field) != -1])
		if info:
			info = ' ' + info
		return f'<Player username={self.username!r}{info}>'

	@classmethod
	def from_packet(cls, packet: Packet):
		"""Reads a Player from a packet.
		:param packet: :class:`aiotfm.Packet` the packet.
		:return: :class:`aiotfm.Player` the player.
		"""
		name = packet.readUTF()
		pid = packet.read32()
		kwargs = {
			'isShaman': packet.readBool(),
			'isDead': packet.read8() > 0, # may be bigger than 1?
			'score': packet.read16(),
			'cheeses': packet.read8(),
			'title': packet.read16(),
			'title_stars': packet.read8() - 1,
			'gender': packet.read8(),
		}
		packet.readUTF() # ???

		look = packet.readUTF()
		packet.readBool() # rasterisation ? wth
		mouseColor = packet.read32()
		shamanColor = packet.read32()
		packet.read32() # ???
		color = packet.read32()
		packet.read8() # respawn id?
		nameColor = -1 if color == 0xFFFFFFFF else color

		kwargs.update({
			'look': look,
			'mouseColor': mouseColor,
			'shamanColor': shamanColor,
			'color': color,
			'nameColor': nameColor
		})

		return cls(name, pid=pid, **kwargs)

	def __str__(self):
		return self.username.capitalize().replace('#0000', '')

	def __eq__(self, other):
		if isinstance(other, str):
			return str(self) == other or self.username.lower() == other.lower()
		if -1 not in [self.id, other.id]:
			return self.id == other.id
		if -1 not in [self.pid, other.pid]:
			return self.pid == other.pid
		return self.username.lower() == other.username.lower()

	@property
	def isGuest(self):
		"""Return True if the player is a guest (Souris)"""
		return self.username.startswith('*')

	@property
	def hasCheese(self):
		"""Return True if the player has cheese."""
		return self.cheeses > 0


class Profile:
	"""Represents a player's profile.

	Attributes
	----------
	username: `str`
		The player's username.
	uid: `int`
		The player's id.
	registration_date: `int`
		The registration timestamp of the player.
	privLevel: `int`
		The privilege level of the player.
	gender: `int`
		Player's gender.
	tribe: `str`
		Player's tribe. Can be `None`.
	soulmate: `str`
		Player's soulmate. Can be `None`.
	title: `int`
		The title above the player's head.
	titles: `set`
		The list of the unlocked titles.
	titles_stars: `dict`
		A dictionary where are stored the number of stars a title has.
	look: `str`
		The player's look.
	level: `int`
		The player's shaman level.
	badges: `dict`
		All badges unlocked by the player and their number.
	stats: `Stats`
		The player's stats.
	equippedOrb: `int`
		The equipped orb of the player.
	orbs: `set`
		The list of unlocked orbs.
	adventurePoints: `int`
		Number of adventure points the player has.
	"""
	def __init__(self, packet: Packet):
		self.username = packet.readUTF()
		self.id = packet.read32()

		self.registration_date = packet.read32()
		self.privLevel = packet.read8()
		self.gender = packet.read8()
		self.tribe = packet.readUTF() or None
		self.soulmate = packet.readUTF() or None
		stats = [packet.read32() for i in range(7)]
		self.title = packet.read16()

		self.titles = set()
		self.titles_stars = {}
		for _ in range(packet.read16()):
			title_id, stars = packet.read16(), packet.read8()
			self.titles.add(title_id)
			if stars > 1:
				self.titles_stars[title_id] = stars

		self.look = packet.readUTF()
		self.level = packet.read16()

		self.badges = {}
		for _ in range(round(packet.read16() / 2)):
			badge, quantity = packet.read16(), packet.read16()
			self.badges[badge] = quantity

		modeStats = []
		for _ in range(packet.read8()):
			modeStats.append((packet.read8(), packet.read32(), packet.read32(), packet.read16()))
		self.stats = Stats(stats, modeStats)

		self.equippedOrb = packet.read8()
		self.orbs = set()
		for _ in range(packet.read8()):
			self.orbs.add(packet.read8())

		packet.readBool()
		self.adventurePoints = packet.read32()


class Stats:
	"""Represents the statistics of a player.

	Attributes
	----------
	normalModeSaves: `int`
		Number of shaman saves in normal mode.
	hardModeSaves: `int`
		Number of shaman saves in hard mode.
	divineModeSaves: `int`
		Number of shaman saves in divine mode.
	shamanCheese: `int`
		Number of cheese personally gathered.
	firsts: `int`
		Number of cheese gathered first.
	gatheredCheese: `int`
		Total number of gathered cheese.
	bootcamps: `int`
		Number of bootcamp.
	modeStats: `list`
		A list of tuples that represents the stats in different mode.
		(id, progress, progressLimit, imageId)
	"""
	def __init__(self, stats, modeStats):
		self.normalModeSaves = stats[0]
		self.hardModeSaves = stats[4]
		self.divineModeSaves = stats[6]
		self.shamanCheese = stats[1]
		self.firsts = stats[2]
		self.gatheredCheese = stats[3]
		self.bootcamps = stats[5]

		self.modeStats = modeStats # id, progress, progressLimit, imageId



from typing import List, Union, Optional

from aiotfm.enums import Game, Permissions
from aiotfm.packet import Packet
from aiotfm.player import Player
from aiotfm.utils import Date


class Tribe:
	"""Represents a tribe.

	Attributes
	----------
	id: :class:`int`
		The tribe's id.
	name: :class:`str`
		The tribe's name.
	welcomeMessage: :class:`str`
		The tribe's welcome message.
	mapcode: :class:`int`
		The tribehouse's mapcode.
	members: :class:`list`
		The members' list of the tribe.
	ranks: :class:`list`
		The ranks' list of the tribe.
	"""
	def __init__(self, packet: Packet):
		self.id: int = packet.read32()
		self.name: str = packet.readUTF()
		self.welcomeMessage: str = packet.readUTF()
		self.mapcode: int = packet.read32()
		self.members: List[Member] = []
		self.ranks: List[Rank] = []

		for i in range(packet.read16()):
			self.members.append(Member(self, packet))

		for i in range(packet.read16()):
			self.ranks.append(Rank.from_packet(i, packet))

	def get_member(self, name: Union[Player, str]) -> Optional['Member']:
		"""Returns a member from it's name or None if not found.
		:param name: :class:`str` or :class:`aiotfm.Player` the name of the member.
		:return: :class:`aiotfm.tribe.Member` or None
		"""
		for m in self.members:
			if name == m.name:
				return m


class Member:
	"""Represents a tribe's member.

	Attributes
	----------
	tribe: :class:`Tribe`
		The member's tribe.
	id: :class:`int`
		The player's id of the member.
	name: :class:`str`
		The username of the member.
	gender: :class:`int`
		The member's gender.
	hasAvatar: :class:`bool`
		True if the player has an avatar
	lastConnection: :class`Date`
		The last connection of the member.
	rank_id: :class:`int`
		The rank's id of the member.
	game: :class:`aiotfm.enums.Game`
		What game the player is playing on
	room: :class:`str`
		The room where the player is.
	rank: :class:`Rank`
		The member's rank.
	online: :class:`bool`
		True if the member is online.
	"""
	def __init__(self, tribe: Tribe, packet: Packet):
		self.tribe: Tribe = tribe
		self.id: int = packet.read32()
		self.name: str = packet.readUTF()
		self.gender: int = packet.read8()
		self.hasAvatar: bool = packet.read32() != 0
		self.lastConnection: Date = Date.fromtimestamp(packet.read32())
		self.rank_id: int = packet.read8()
		self.game: Game = Game(packet.read32())
		self.room: str = packet.readUTF()

	@property
	def rank(self) -> 'Rank':
		"""return the :class:`Rank` of the member."""
		return self.tribe.ranks[self.rank_id]

	@property
	def avatar(self) -> str:
		"""Return the player's avatar's url."""
		if self.hasAvatar:
			return f'https://avatars.atelier801.com/{self.id % 10000}/{self.id}.jpg'

		# default avatar
		return 'https://avatars.atelier801.com/0/0.jpg'

	@property
	def online(self) -> bool:
		"""return True if the member is online."""
		return self.game != Game.INVALID


class Rank:
	"""Represents a tribe's rank.

	Attributes
	----------
	id: :class:`int`
		The rank's id.
	name: :class:`str`
		The rank's name.
	perm: :class:`int`
		The rank's permissions.
	"""
	def __init__(self, id_: int, name: str, perm: int):
		self.id: int = id_
		self.name: str = name
		self.perm: int = perm

	@property
	def isLeader(self) -> bool:
		"""True if it's the tribe's leader's rank."""
		return bool(self.perm & Permissions.IS_LEADER)

	@property
	def canChangeGreetingMessage(self) -> bool:
		"""True if it has the permission to change the greeting message."""
		return bool(self.perm & Permissions.CAN_CHANGE_GREETING_MESSAGE)

	@property
	def canEditRanks(self) -> bool:
		"""True if it has the permission to edit ranks."""
		return bool(self.perm & Permissions.CAN_EDIT_RANKS)

	@property
	def canChangeMembersRanks(self) -> bool:
		"""True if it has the permission to change members' rank."""
		return bool(self.perm & Permissions.CAN_CHANGE_MEMBERS_RANKS)

	@property
	def canInvite(self) -> bool:
		"""True if it has the permission to invite someone to the tribe."""
		return bool(self.perm & Permissions.CAN_INVITE)

	@property
	def canExclude(self) -> bool:
		"""True if it has the permission to exclude someone of the tribe."""
		return bool(self.perm & Permissions.CAN_EXCLUDE)

	@property
	def canPlayMusic(self) -> bool:
		"""True if it has the permission to play music inside the tribe's house."""
		return bool(self.perm & Permissions.CAN_PLAY_MUSIC)

	@property
	def canChangeTribeHouseMap(self) -> bool:
		"""True if it has the permission to change the tribe's house's map."""
		return bool(self.perm & Permissions.CAN_CHANGE_TRIBE_HOUSE_MAP)

	@property
	def canLoadMap(self) -> bool:
		"""True if it has the permission to load maps inside the tribe's house."""
		return bool(self.perm & Permissions.CAN_LOAD_MAP)

	@property
	def canLoadLua(self) -> bool:
		"""True if it has the permission to load Lua inside the tribe's house."""
		return bool(self.perm & Permissions.CAN_LOAD_LUA)

	@property
	def canManageForum(self) -> bool:
		"""True if it has the permission to mange the tribe's forum."""
		return bool(self.perm & Permissions.CAN_MANAGE_FORUM)

	@classmethod
	def from_packet(cls, id_: int, packet: Packet):
		"""Reads a Tribe from a packet.
		:param id: :class:`int` the tribe's id.
		:param packet: :class:`aiotfm.Packet`"""
		return cls(id_, packet.readUTF(), packet.read32())



from aiotfm.enums import TradeState


class AiotfmException(Exception):
	"""Base exception class for aiotfm"""


class LoginError(AiotfmException):
	"""Exception thrown when the login failed."""
	def __init__(self, code: int):
		self.code = code
		super().__init__('Login Failed ! Error code: {.code}.'.format(self))


class AlreadyConnected(LoginError):
	"""Exception thrown when the account provided is already connected."""
	def __init__(self):
		super().__init__(1)


class IncorrectPassword(LoginError):
	"""Exception thrown when trying to connect with a wrong password."""
	def __init__(self):
		super().__init__(2)


class InvalidEvent(AiotfmException):
	"""Exception thrown when you added an invalid event to the client.

	An event is valid only if its name begin by 'on_' and it is coroutine.
	"""


class ServerUnreachable(AiotfmException):
	"""Exception thrown when the Client can't connect to the server."""


class ConnectionClosed(AiotfmException):
	"""Exception thrown when one of the connection closes."""


class InvalidSocketData(AiotfmException):
	"""Exception thrown when a socket receive an invalid data."""


class EndpointError(AiotfmException):
	"""Exception thrown when the endpoint sends an abnormal response."""


class InternalError(EndpointError):
	"""Exception thrown when the endpoint got an internal error."""


class MaintenanceError(EndpointError):
	"""Exception thrown when the endpoint thinks there is a maintenance."""


class InvalidLocale(AiotfmException):
	"""Exception thrown when you try to load an inexistent locale."""


class PacketError(AiotfmException):
	"""Exception thrown when a packet encounter a problem."""


class PacketTooLarge(PacketError):
	"""Exception thrown when a packet is too large to be exported."""


class XXTEAError(PacketError):
	"""Exception thrown when the XXTEA algorithm failed."""


class XXTEAInvalidPacket(XXTEAError):
	"""Exception thrown when you try to cipher an empty Packet."""


class XXTEAInvalidKeys(XXTEAError):
	"""Exception thrown when you try to cipher a packet with an invalid key."""


class CommunityPlatformError(AiotfmException):
	"""Exception thrown when the community platform send an error code."""
	def __init__(self, category: int, code: int):
		super().__init__('Internal error: {}-{}'.format(category, code))
		self.category = category
		self.code = code


class TradeOnWrongState(AiotfmException):
	"""Exception thrown when the client try an impossible action on trade due to its state."""
	def __init__(self, action: str, state: TradeState):
		super().__init__(f'Can not {action} when the trade is {state}.')
		self.action: str = action
		self.state: TradeState = state


class InvalidAccountError(AiotfmException):
	"""Exception thrown when a server action does not find a user."""
	def __init__(self, player: str):
		super().__init__(f'The account {player} does not exist.')
		self.player: str = player


class FriendLimitError(AiotfmException):
	"""Exception thrown when your friend list is full."""
	def __init__(self):
		super().__init__('You can\'t add more friends.')


class CantFriendPlayerError(AiotfmException):
	"""Exception thrown when the server does not let you friend a player."""
	def __init__(self, player: str):
		super().__init__(f'You can\'t friend the player {player}')
		self.player: str = player



from aiotfm import enums, errors, utils
from aiotfm.__version__ import __author__, __credits__, __description__, __license__, __title__, __url__, __version__
from aiotfm.client import Client
from aiotfm.connection import Connection
from aiotfm.enums import GameMode
from aiotfm.packet import Packet
from aiotfm.player import Player, Profile, Stats
from aiotfm.tribe import Member, Rank, Tribe

__all__ = [
	'__author__', '__credits__', '__description__', '__license__', '__title__', '__url__',
	'__version__', 'enums', 'errors', 'utils', 'Client', 'Connection', 'Member', 'Packet',
	'Player', 'Profile', 'Rank', 'Stats', 'Tribe', 'GameMode'
]



import re
import zlib

import aiohttp

from aiotfm.errors import InvalidLocale


class Translation:
	"""Represents a translation item of the game.

	Parameters
	----------
	key: :class:`str`
		The translation's key.
	value: :class:`str`
		The translated text.

	Attributes
	----------
	key: :class:`str`
		The translation's key.
	value: :class:`str`
		The translated text.
	"""
	def __init__(self, key, value):
		self.key = key
		self.value = value

	def __str__(self):
		return self.value

	def __repr__(self):
		return f'{self.key}={self.value}'

	def format(self, *args):
		"""Format the translation value, replacing %N place holders by arguments.

		:params: the values to inject.
		:return: :class:`str` the formatted output."""
		def repl(match):
			index = int(match.group(1))
			if index > len(args) or index == 0:
				return match.group(0)
			return str(args[index - 1])

		return re.sub(r'%(\d+)', repl, self.value)


class Locale:
	"""Represents the locale file of the game.

	Parameters
	----------
	locale: :class:`str`
		The locale name.

	Attributes
	----------
	locales: :class:`dict`[:class:`aiotfm.locale.Locale`]
		Cached locales.
	locale: :class:`str`
		The locale name.
	"""
	BASE_URL = 'http://transformice.com/langues/tfz_{}'

	def __init__(self, locale='en'):
		self._locale = locale
		self.locales = {}

	def __getitem__(self, key):
		"""Return the translation of a key.

		:param key: :class:`str` the translation's key
		:return: :class:`aiotfm.locale.Translation`
		"""
		table = self.locales.get(self.locale, {})
		value = table.get(key[1:] if key[0] == '$' else key, None)

		if value is None:
			return Translation(key, key)

		return Translation(key, value)

	@property
	def locale(self):
		"""The active locale."""
		return self._locale

	async def reload(self, locale=None):
		"""|coro|
		Reload a locale and set it as active.

		:param locale: Optional[:class:`str`] The locale name you want to reload.
			If it's None, the active locale is used instead.
		"""
		if locale is None:
			locale = self._locale

		# Deletes the locale then load it
		del self.locales[locale]
		await self.load(locale)

	async def load(self, locale=None):
		"""|coro|
		Load a locale and set it as active.

		:param locale: Optional[:class:`str`] The locale name you want to load.
			If it's None, the active locale is used instead.
		"""
		if locale is not None:
			self._locale = locale

		# Check if the locale is cached
		if self._locale in self.locales:
			return

		# Download the locale
		async with aiohttp.ClientSession() as session:
			async with session.get(self.BASE_URL.format(self._locale)) as r:
				if r.status == 404:
					raise InvalidLocale()

				# Decompress the file and parse it
				content = zlib.decompress(await r.read()).decode('utf-8')
				table = {k: v for k, v in (t.split('=', 1) for t in content.split('\n-\n') if t)}
				# Add the translation table
				self.locales[self._locale] = table



import asyncio
import logging
from asyncio import AbstractEventLoop, Protocol, BaseTransport, Transport
from typing import Optional, Tuple

import aiotfm  # circular import, don't `import from`
from aiotfm.errors import AiotfmException

logger = logging.getLogger('aiotfm')


class TFMProtocol(Protocol):
	def __init__(self, conn):
		self.buffer: bytearray = bytearray()
		self.client: aiotfm.Client = conn.client
		self.connection: Connection = conn
		self.length: int = 0

	def data_received(self, data: bytes):
		self.buffer.extend(data)

		while len(self.buffer) > self.length:
			if self.length == 0:
				for i in range(5):
					byte = self.buffer.pop(0)
					self.length |= (byte & 127) << (i * 7)

					if not byte & 0x80:
						break
				else:
					raise Exception("wtf") # pragma: no cover

			if len(self.buffer) >= self.length:
				self.client.data_received(self.buffer[:self.length], self.connection)
				del self.buffer[:self.length]
				self.length = 0

	def connection_made(self, transport: BaseTransport):
		# :desc: Called when a connection has been successfully made with the server.
		# :param connection: :class:`Connection` the connection that has been made.
		self.connection.open = True
		self.client.dispatch('connection_made', self.connection)

	def connection_lost(self, exc: Optional[Exception] = None):
		self.connection.open = False

		if exc is None:
			logger.info('Connection %s has been lost.', self.connection.name)
		else:
			logger.error('Connection %s has been lost. Reason:', self.connection.name, exc_info=exc)
			# :desc: Called when a connection has been lost due to an error.
			# :param connection: :class:`Connection` the connection that has been lost.
			# :param exception: :class:`Exception` the error which occurred.
			self.client.dispatch('connection_error', self.connection, exc)

		if self.connection.name == "main" and not self.client._close_event.done():
			self.client._close_event.set_result(('connection_lost', 10, None))


class Connection:
	"""Represents the connection between the client and the host."""
	PROTOCOL = TFMProtocol

	def __init__(self, name: str, client: 'aiotfm.Client', loop: AbstractEventLoop):
		self.name: str = name
		self.client: aiotfm.Client = client
		self.loop: AbstractEventLoop = loop

		self.address: Tuple[str, int] = None
		self.protocol: Protocol = None
		self.transport: Transport = None

		self.fingerprint: int = 0
		self.open: bool = False

	def __bool__(self):
		return self.open

	def _factory(self):
		return Connection.PROTOCOL(self)

	async def connect(self, host: str, port: int):
		"""|coro|
		Connect the client to the host:port
		"""
		self.address = (host, port)
		self.transport, self.protocol = await asyncio.wait_for(self.loop.create_connection(self._factory, host, port), 3)

	async def send(self, packet: 'aiotfm.Packet', cipher: bool = False):
		"""|coro|
		Send a packet to the socket

		:param packet: :class:`aiotfm.Packet` the packet to send.
		:param cipher: :class:`bool` whether or not the packet should be ciphered before sending it.
		"""
		if not self.open:
			raise AiotfmException('Cannot send a packet to a closed Connection.')

		if not self.client.bot_role and cipher:
			packet.xor_cipher(self.client.keys.msg, self.fingerprint)

		self.transport.write(packet.export(self.fingerprint))
		self.fingerprint = (self.fingerprint + 1) % 100

	def close(self):
		"""Closes the connection."""
		self.open = False
		if self.transport is not None and not self.transport.is_closing():
			self.transport.write_eof()
			self.transport.close()

	def abort(self):
		"""Abort the connection."""
		self.transport.abort()
		self.close()



from typing import List, Optional, Union, Iterable

import aiotfm
from aiotfm.enums import Game
from aiotfm.errors import CantFriendPlayerError, CommunityPlatformError, FriendLimitError, InvalidAccountError
from aiotfm.packet import Packet
from aiotfm.player import Player
from aiotfm.utils import Date


class FriendList:
	"""Represents a friend list.

	Attributes
	----------
	soulmate: Optional[:class:`aiotfm.friend.Friend`]
		Your soulmate, if you have one.
	friends: :class:`list`
		Your friends.
	"""
	def __init__(self, client: 'aiotfm.Client', packet: Packet):
		self.friends: List[Friend] = []
		self.soulmate: Optional[Friend] = None

		soulmate = Friend(self, packet, True)
		if soulmate.id != 0:
			self.soulmate = soulmate
			self.friends.append(soulmate)

		for i in range(packet.read16()):
			self.friends.append(Friend(self, packet))

		self._client: aiotfm.Client = client

	def __iter__(self) -> Iterable['Friend']:
		return iter(self.friends)

	def get_friend(self, search: Union[Player, str, int]) -> Optional['Friend']:
		"""Returns a friend from their name (or id) or None if not found.
		:param search: :class:`str` or :class:`aiotfm.Player` or :class:`int` search query
		:return: :class:`aiotfm.friend.Friend` or None
		"""
		if isinstance(search, int): # Search by id
			for f in self.friends:
				if search == f.id:
					return f

		else:
			if isinstance(search, Player):
				search = search.username
			search = search.lower()

			for f in self.friends:
				if search == f.name:
					return f

	async def remove(self, friend: Union[Player, 'Friend', str]):
		"""|coro|
		Remove a friend. If they're your soulmate, divorce them.
		:param friend: :class:`str` or :class:`aiotfm.Player` or :class:`aiotfm.friend.Friend`
		"""
		if not isinstance(friend, Friend):
			friend = self.get_friend(friend)

			if friend is None:
				return

		if friend.isSoulmate:
			sid = await self._client.sendCP(26, Packet())
			result, error = 27, 36
		else:
			sid = await self._client.sendCP(20, Packet().writeString(friend.name.lower()))
			result, error = 21, 30

		def is_deletion(tc, packet):
			return tc == result and packet.read32() == sid

		tc, packet = await self._client.wait_for('on_raw_cp', is_deletion, timeout=5)
		result = packet.read8()

		if result != 1:
			raise CommunityPlatformError(error, result)

		if friend not in self.friends:
			return

		if friend == self.soulmate:
			self.soulmate = None
		self.friends.remove(friend)

	async def add(self, name: Union[Player, str]) -> Optional['Friend']:
		"""|coro|
		Add a friend.
		:param name: :class:`str` or :class:`aiotfm.Player`
		:return: :class:`aiotfm.friend.Friend` or None
		"""
		if isinstance(name, Player):
			name = name.username

		friend = self.get_friend(name)
		if friend is not None:
			return friend

		sid = await self._client.sendCP(18, Packet().writeString(name))

		def is_addition(tc, packet):
			return tc == 19 and packet.read32() == sid

		tc, packet = await self._client.wait_for('on_raw_cp', is_addition, timeout=5)
		result = packet.read8()

		if result == 12:
			raise InvalidAccountError(name)

		elif result == 7:
			raise FriendLimitError()

		elif result == 4 or result == 15:
			raise CantFriendPlayerError(name)

		elif result != 1:
			raise CommunityPlatformError(28, result)

		return self.get_friend(name)


class Friend:
	"""Represents a player in friend list, if you have one :feels:

	Attributes
	----------
	id: :class:`int`
		The player's id
	name: :class:`str`
		The player's username.
	gender: :class:`int`
		The player's gender.
	isSoulmate: :class:`bool`
		True if the player is your soulmate
	isAddedBack: :class:`bool`
		True if the the player also added you to their friend list
	isOnline: :class:`bool`
		True if the player is online
	hasAvatar: :class:`bool`
		True if the player has an avatar
	game: :class:`aiotfm.enums.Game`
		What game the player is playing on
	roomName: :class:`str`
		The player's room name, empty string if isAddedBack is False
	lastConnection: :class:`Date`
		The last connection of the player
	"""

	def __init__(self, flist: FriendList, packet: Packet, isSoulmate: bool = False):
		self.id: int = packet.read32()
		self.name: str = packet.readUTF()
		self.gender: int = packet.read8()
		self.hasAvatar: bool = packet.read32() != 0
		self.isSoulmate: bool = isSoulmate
		self.isAddedBack: bool = packet.readBool()
		self.isConnected: bool = packet.readBool()
		self.game: Game = Game(packet.read32())
		self.roomName: str = packet.readUTF()
		self.lastConnection: Date = Date.fromtimestamp(packet.read32())

		self._flist: FriendList = flist

	@property
	def avatar(self) -> str:
		"""Return the player's avatar's url."""
		if self.hasAvatar:
			return f'https://avatars.atelier801.com/{self.id % 10000}/{self.id}.jpg'

		# default avatar
		return 'https://avatars.atelier801.com/0/0.jpg'

	def remove(self):
		"""|coro|
		Remove this friend. If they're your soulmate, divorce them.
		"""
		return self._flist.remove(self)



import aiohttp

from aiotfm import __version__
from aiotfm.errors import EndpointError, InternalError, MaintenanceError


class Keys:
	"""Represents the keys used by the client to communicate to the server."""
	def __init__(self, **keys):
		self.auth = keys.pop('auth', 0)
		self.connection = keys.pop('connection', '')
		self.identification = keys.pop('identification', [])
		self.msg = [k & 0xff for k in keys.pop('msg', [])]
		self.packet = keys.pop('packet', [])
		self.version = keys.pop('version', 0)
		self.server_ip = keys.pop('ip', '193.70.81.30')
		self.server_ports = keys.pop('ports', [11801, 12801, 13801, 14801])
		self.kwargs = keys


async def request_api(url):
	"""|coro|
	Sends a GET request to the specified URL, expecting a JSON response.
	If the client is unable to receive a proper JSON response,
	an empty dictionary is returned.

	:param url: :class:`str` The URL to send the request to.
	"""
	headers = {"User-Agent": f"Mozilla/5.0 aiotfm/{__version__}"}

	try:
		async with aiohttp.ClientSession() as session:
			async with session.get(url, headers=headers) as resp:
				return await resp.json()
	except aiohttp.ClientError:
		return {}


async def get_ip():
	"""|coro|
	Fetch the game IP and ports, useful for bots with the official role.
	"""
	url = 'https://cheese.formice.com/api/tfm/ip'
	data = await request_api(url)

	if not len(data):
		# Empty dictionary, request failed, let's use default server IP
		success = True
	else:
		success = data.pop('success', False)
		error = data.pop('error', '').capitalize()
		description = data.pop('description', 'No description was provided.')

	if not success:
		if error == 'Maintenance':
			raise MaintenanceError('The game is under maintenance.')

		if error == 'Internal':
			raise InternalError(description)

		raise EndpointError(f'{error}: {description}')

	return Keys(version=666, **data.get('server', {}))


async def get_keys(tfm_id, token):
	"""|coro|
	Fetch the keys required to log into the game.

	:param tfm_id: :class:`int` your Transformice user id.
	:param token: :class:`str` your api token.
	"""
	url = f'https://api.tocuto.tk/tfm/get/keys/{tfm_id}/{token}'
	data = await request_api(url)

	success = data.pop('success', False)
	error = data.pop('error', '').capitalize()
	description = data.pop('description', 'No description was provided.')

	if not success:
		if error == 'Maintenance':
			raise MaintenanceError('The game is under maintenance.')

		if error == 'Internal':
			raise InternalError(description)

		raise EndpointError(f'{error}: {description}')

	keys = Keys(**data.get('server', {}), **data.get('keys', {}))
	if len(keys.packet) > 0 and len(keys.identification) > 0 and len(keys.msg) > 0 and keys.version != 0:
		return keys

	raise EndpointError('Something went wrong: A key is empty ! {}'.format(data))



from aiotfm.utils.date import Date
from aiotfm.utils.get_keys import Keys, get_ip, get_keys
from aiotfm.utils.locale import Locale, Translation
from aiotfm.utils.shakikoo import shakikoo

__all__ = ['shakikoo', 'Date', 'get_keys', 'get_ip', 'Translation', 'Locale', 'Keys']



import asyncio
import logging
import random
import warnings
from typing import AnyStr, ByteString, Callable, List, Optional, Union

from aiotfm.connection import Connection
from aiotfm.enums import Community, GameMode, TradeError
from aiotfm.errors import AiotfmException, AlreadyConnected, CommunityPlatformError, \
	IncorrectPassword, InvalidEvent, LoginError, MaintenanceError, ServerUnreachable
from aiotfm.friend import Friend, FriendList
from aiotfm.inventory import Inventory, InventoryItem, Trade
from aiotfm.message import Channel, ChannelMessage, Message, Whisper
from aiotfm.packet import Packet
from aiotfm.player import Player, Profile
from aiotfm.room import Room, RoomList
from aiotfm.shop import Shop
from aiotfm.tribe import Tribe
from aiotfm.utils import Keys, Locale, get_ip, get_keys, shakikoo

logger = logging.getLogger('aiotfm')


class Client:
	"""Represents a client that connects to Transformice.
	Two argument can be passed to the :class:`Client`.

	.. _event loop: https://docs.python.org/3/library/asyncio-eventloops.html

	Parameters
	----------
	community: Optional[:class:`int`]
		Defines the community of the client. Defaults to 0 (EN community).
	auto_restart: Optional[:class:`bool`]
		Whether the client should automatically restart on error. Defaults to False.
	bot_role: Optional[:class:`bool`]
		Whether the has the game's special role bot or not.
		Avoids using the api endpoint and gives more stability.
	loop: Optional[event loop]
		The `event loop`_ to use for asynchronous operations. If ``None`` is passed (defaults),
		the event loop used will be ``asyncio.get_event_loop()``.

	max_retries: Optional[:class:`int`]
		The maximum number of retries the client should attempt while connecting to the game.

	Attributes
	----------
	username: Optional[:class:`str`]
		The bot's username received from the server. Might be None if the bot didn't log in yet.
	room: Optional[:class:`aiotfm.room.Room`]
		The bot's room. Might be None if the bot didn't log in yet or couldn't join any room yet.
	trade: Optional[:class:`aiotfm.inventory.Trade`]
		The current trade that's going on (i.e: both traders accepted it).
	trades: :class:`list`[:class:`aiotfm.inventory.Trade`]
		All the trades that the bot participates. Most of them might be invitations only.
	inventory: Optional[:class:`aiotfm.inventory.Inventory`]
		The bot's inventory. Might be None if the bot didn't log in yet or it didn't receive
		anything.
	locale: :class:`aiotfm.locale.Locale`
		The bot's locale (translations).
	friends: Optional[:class:`aiotfm.friends.FriendList`]
		The bot's friend list
	"""
	LOG_UNHANDLED_PACKETS = False

	def __init__(
		self,
		community: Union[Community, int] = Community.en,
		auto_restart: bool = False,
		bot_role: bool = False,
		loop: Optional[asyncio.AbstractEventLoop] = None,
		max_retries: int = 6
	):
		self.loop: asyncio.AbstractEventLoop = loop or asyncio.get_event_loop()

		self.main: Connection = Connection('main', self, self.loop)
		self.bulle: Connection = None

		self._waiters: dict = {}
		self._close_event: asyncio.Future = None
		self._sequenceId: int = 0
		self._channels: List[Channel] = []
		self._restarting: bool = False
		self._closed: bool = False
		self._logged: bool = False
		self._max_retries: int = max_retries

		self.room: Room = None
		self.trade: Trade = None
		self.trades: dict = {}
		self.inventory: Inventory = None

		self.username: str = None
		self.locale: Locale = Locale()
		self.community: Community = Community(community)

		self.friends: FriendList = None

		self.keys: Keys = None
		self.authkey: int = 0

		self.auto_restart: bool = auto_restart
		self.api_tfmid: int = None
		self.api_token: str = None
		self.bot_role: bool = bot_role

	@property
	def restarting(self) -> bool:
		return self._restarting

	@property
	def closed(self) -> bool:
		return self._closed

	def _backoff(self, n: int) -> float:
		"""Returns the numbers of seconds to wait until the n-th connection attempt. Capped at 10 minutes."""
		return random.uniform(20, 30 * 2 ** min(n, 5))

	def data_received(self, data: bytes, connection: Connection):
		"""|coro|
		Dispatches the received data.

		:param data: :class:`bytes` the received data.
		:param connection: :class:`aiotfm.Connection` the connection that received
			the data.
		"""
		# :desc: Called when a socket receives a packet. Does not interfere
		# with :meth:`Client.handle_packet`.
		# :param connection: :class:`aiotfm.Connection` the connection that received
		# the packet.
		# :param packet: :class:`aiotfm.Packet` a copy of the packet.
		self.dispatch('raw_socket', connection, Packet(data))
		self.loop.create_task(self.handle_packet(connection, Packet(data)))

	async def handle_packet(self, connection: Connection, packet: Packet) -> bool:
		"""|coro|
		Handles the known packets and dispatches events.
		Subclasses should handle only the unhandled packets from this method.

		Example: ::
			class Bot(aiotfm.Client):
				async def handle_packet(self, conn, packet):
					handled = await super().handle_packet(conn, packet.copy())

					if not handled:
						# Handle here the unhandled packets.
						pass

		:param connection: :class:`aiotfm.Connection` the connection that received
			the packet.
		:param packet: :class:`aiotfm.Packet` the packet.
		:return: True if the packet got handled, False otherwise.
		"""
		CCC = packet.readCode()
		if CCC == (1, 1): # Old packets
			oldCCC, *data = packet.readString().split(b'\x01')
			data = list(map(bytes.decode, data))
			oldCCC = tuple(oldCCC[:2])

			# :desc: Called when an old packet is received. Does not interfere
			# with :meth:`Client.handle_old_packet`.
			# :param connection: :class:`aiotfm.Connection` the connection that received
			# the packet.
			# :param oldCCC: :class:`tuple` the packet identifiers on the old protocol.
			# :param data: :class:`list` the packet data.
			self.dispatch('old_packet', connection, oldCCC, data)
			return await self.handle_old_packet(connection, oldCCC, data)

		if CCC == (5, 21): # Joined room
			self.room = Room(official=packet.readBool(), name=packet.readUTF())

			# :desc: Called when the client has joined a room.
			# :param room: :class:`aiotfm.room.Room` the room the client has entered.
			self.dispatch('joined_room', self.room)

		elif CCC == (5, 39): # Password required for the room

			# :desc: Called when a password is required to enter a room
			# :param room: :class:`aiotfm.room.Room` the room the server is asking for a password.
			self.dispatch('room_password', Room(packet.readUTF()))

		elif CCC == (6, 6): # Room message
			username = packet.readUTF()
			message = packet.readUTF()
			player = self.room.get_player(username=username)

			if player is None:
				player = Player(username)

			# :desc: Called when the client receives a message from the room.
			# :param message: :class:`aiotfm.message.Message` the message.
			self.dispatch('room_message', Message(player, message, self))

		elif CCC == (6, 20): # Server message
			packet.readBool() # if False then the message will appear in the #Server channel
			t_key = packet.readUTF()
			t_args = [packet.readUTF() for i in range(packet.read8())]

			# :desc: Called when the client receives a message from the server that needs to be translated.
			# :param message: :class:`aiotfm.locale.Translation` the message translated with the
			# current locale.
			# :param *args: a list of string used as replacement inside the message.
			self.dispatch('server_message', self.locale[t_key], *t_args)

		elif CCC == (8, 1): # Play emote
			player = self.room.get_player(pid=packet.read32())
			emote = packet.read8()
			flag = packet.readUTF() if emote == 10 else ''

			# :desc: Called when a player plays an emote.
			# :param player: :class:`aiotfm.Player` the player.
			# :param emote: :class:`int` the emote's id.
			# :param flag: :class:`str` the flag's id.
			self.dispatch('emote', player, emote, flag)

		elif CCC == (8, 5): # Show emoji
			player = self.room.get_player(pid=packet.read32())
			emoji = packet.read8()

			# :desc: Called when a player is showing an emoji above its head.
			# :param player: :class:`aiotfm.Player` the player.
			# :param emoji: :class:`int` the emoji's id.
			self.dispatch('emoji', player, emoji)

		elif CCC == (8, 6): # Player won
			packet.read8()
			player = self.room.get_player(pid=packet.read32())
			player.score = packet.read16()
			order = packet.read8()
			player_time = packet.read16() / 100

			# :desc: Called when a player get the cheese to the hole.
			# :param player: :class:`aiotfm.Player` the player.
			# :param order: :class:`int` the order of the player in the hole.
			# :param player_time: :class:`float` player's time in the hole in seconds.
			self.dispatch('player_won', player, order, player_time)

		elif CCC == (8, 16): # Profile
			# :desc: Called when the client receives the result of a /profile command.
			# :param profile: :class:`aiotfm.player.Profile` the profile.
			self.dispatch('profile', Profile(packet))

		elif CCC == (8, 20): # Shop
			# :desc: Called when the client receives the content of the shop.
			# :param shop: :class:`aiotfm.shop.Shop` the shop.
			self.dispatch('shop', Shop(packet))

		elif CCC == (8, 22): # Skills
			skills = {}
			for _ in range(packet.read8()):
				key, value = packet.read8(), packet.read8()
				skills[key] = value

			# :desc: Called when the client receives its skill tree.
			# :param skills: :class:`dict` the skills.
			self.dispatch('skills', skills)

		elif CCC == (16, 2): # Tribe invitation received
			author = packet.readUTF()
			tribe = packet.readUTF()

			# :desc: Called when the client receives an invitation to a tribe. (/inv)
			# :param author: :class:`str` the player that invited you.
			# :param tribe: :class:`str` the tribe.
			self.dispatch('tribe_inv', author, tribe)

		elif CCC == (26, 2): # Logged in successfully
			player_id = packet.read32()
			self.username = username = packet.readUTF()
			played_time = packet.read32()
			community = Community(packet.read8())
			pid = packet.read32()

			# :desc: Called when the client successfully logged in.
			# :param uid: :class:`int` the client's unique id.
			# :param username: :class:`str` the client's username.
			# :param played_time: :class:`int` the total number of minutes the client has played.
			# :param community: :class:`aiotfm.enums.Community` the community the client has connected to.
			# :param pid: :class:`int` the client's player id.
			self.dispatch('logged', player_id, username, played_time, community, pid)

		elif CCC == (26, 3): # Handshake OK
			online_players = packet.read32()
			language = packet.readUTF()
			country = packet.readUTF()
			self.authkey = packet.read32()
			self._logged = False

			os_info = Packet.new(28, 17).writeString('en').writeString('Linux')
			os_info.writeString('LNX 29,0,0,140').write8(0)

			await connection.send(os_info)

			# :desc: Called when the client can login through the game.
			# :param online_players: :class:`int` the number of player connected to the game.
			# :param language: :class:`str` the language the server is suggesting.
			# :param country: :class:`str` the country detected from your ip.
			self.dispatch('login_ready', online_players, language, country)

		elif CCC == (26, 12): # Login result
			self._logged = False
			# :desc: Called when the client failed logging.
			# :param code: :class:`int` the error code.
			# :param error1: :class:`str` error messages.
			# :param error2: :class:`str` error messages.
			self.dispatch('login_result', packet.read8(), packet.readUTF(), packet.readUTF())

		elif CCC == (26, 25): # Ping
			# :desc: Called when the client receives the ping response from the server.
			self.dispatch('ping')

		elif CCC == (26, 35): # Room list
			roomlist = RoomList.from_packet(packet)
			# :desc: Dispatched when the client receives the room list
			self.dispatch('room_list', roomlist)

		elif CCC == (28, 6): # Server ping
			await connection.send(Packet.new(28, 6).write8(packet.read8()))

		elif CCC == (29, 6): # Lua logs
			# :desc: Called when the client receives lua logs from #Lua.
			# :param log: :class:`str` a log message.
			self.dispatch('lua_log', packet.readUTF())

		elif CCC == (31, 1): # Inventory data
			self.inventory = Inventory.from_packet(packet)
			self.inventory.client = self

			# :desc: Called when the client receives its inventory's content.
			# :param inventory: :class:`aiotfm.inventory.Inventory` the client's inventory.
			self.dispatch('inventory_update', self.inventory)

		elif CCC == (31, 2): # Update inventory item
			item_id = packet.read16()
			quantity = packet.read8()

			if item_id in self.inventory.items:
				item = self.inventory.items[item_id]
				previous = item.quantity
				item.quantity = quantity

				# :desc: Called when the quantity of an item has been updated.
				# :param item: :class:`aiotfm.inventory.InventoryItem` the new item.
				# :param previous: :class:`int` the previous quantity.
				self.dispatch('item_update', item, previous)

			else:
				item = InventoryItem(item_id=item_id, quantity=quantity)
				self.inventory.items[item.id] = item

				# :desc: Called when the client receives a new item in its inventory.
				# :param item: :class:`aiotfm.inventory.InventoryItem` the new item.
				self.dispatch('new_item', item)

		elif CCC == (31, 5): # Trade invite
			pid = packet.read32()

			self.trades[pid] = Trade(self, self.room.get_player(pid=pid))

			# :desc: Called when received an invitation to trade.
			# :param trade: :class:`aiotfm.inventory.Trade` the trade object.
			self.dispatch('trade_invite', self.trades[pid])

		elif CCC == (31, 6): # Trade error
			name = packet.readUTF().lower()
			error = packet.read8()

			if name == self.username.lower():
				trade = self.trade
			else:
				for t in self.trades.values():
					if t.trader.lower() == name:
						trade = t
						break

			# :desc: Called when an error occurred with a trade.
			# :param trade: :class:`aiotfm.inventory.Trade` the trade that failed.
			# :param error: :class:`aiotfm.enums.TradeError` the error.
			self.dispatch('trade_error', trade, TradeError(error))
			trade._close()

		elif CCC == (31, 7): # Trade start
			pid = packet.read32()
			trade = self.trades.get(pid)

			if trade is None:
				raise AiotfmException(f'Cannot find the trade from pid {pid}.')

			trade._start()
			self.trade = trade

			# :desc: Called when a trade starts. You can access the trade object with `Client.trade`.
			self.dispatch('trade_start')

		elif CCC == (31, 8): # Trade items
			export = packet.readBool()
			id_ = packet.read16()
			quantity = (1 if packet.readBool() else -1) * packet.read8()

			items = self.trade.exports if export else self.trade.imports
			items.add(id_, quantity)

			trader = self if export else self.trade.trader
			self.trade.locked = [False, False]

			# :desc: Called when an item has been added/removed from the current trade.
			# :param trader: :class:`aiotfm.Player` the player that triggered the event.
			# :param id: :class:`int` the item's id.
			# :param quantity: :class:`int` the quantity added/removed. Can be negative.
			# :param item: :class:`aiotfm.inventory.InventoryItem` the item after the change.
			self.dispatch('trade_item_change', trader, id_, quantity, items.get(id_))

		elif CCC == (31, 9): # Trade lock
			index = packet.read8()
			locked = packet.readBool()
			if index > 1:
				self.trade.locked = [locked, locked]
				who = "both"
			else:
				self.trade.locked[index] = locked
				who = self.trade.trader if index == 0 else self

			# :desc: Called when the trade got (un)locked.
			# :param who: :class:`aiotfm.Player` the player that triggered the event.
			# :param locked: :class:`bool` either the trade got locked or unlocked.
			self.dispatch('trade_lock', who, locked)

		elif CCC == (31, 10): # Trade complete
			trade = self.trade
			self.trade._close(succeed=True)

		elif CCC == (44, 1): # Bulle switching
			timestamp = packet.read32()
			uid = packet.read32()
			pid = packet.read32()
			bulle_ip = packet.readUTF()
			ports = packet.readUTF().split('-')

			if self.bulle is not None:
				self.bulle.close()

			self.bulle = Connection('bulle', self, self.loop)
			await self.bulle.connect(bulle_ip, int(random.choice(ports)))
			await self.bulle.send(Packet.new(44, 1).write32(timestamp).write32(uid).write32(pid))

		elif CCC == (44, 22): # Fingerprint offset changed
			connection.fingerprint = packet.read8()

		elif CCC == (60, 3): # Community platform
			TC = packet.read16()

			# :desc: Called when the client receives a packet from the community platform.
			# :param TC: :class:`int` the packet's code.
			# :param packet: :class:`aiotfm.Packet` the packet.
			self.dispatch('raw_cp', TC, packet.copy(copy_pos=True))

			if TC == 3: # Connected to the community platform
				await self.sendCP(28) # Request friend list

				# :desc: Called when the client is successfully connected to the community platform.
				self.dispatch('ready')

			elif TC == 32: # Friend connected
				if self.friends is None:
					return True

				friend = self.friends.get_friend(packet.readUTF())
				friend.isConnected = True

				# :desc: Called when a friend connects to the game (not entirely fetched)
				# :param friend: :class:`aiotfm.friend.Friend` friend after this update
				self.dispatch('friend_connected', friend)

			elif TC == 33: # Friend disconnected
				if self.friends is None:
					return True

				friend = self.friends.get_friend(packet.readUTF())
				friend.isConnected = False

				# :desc: Called when a friend disconnects from the game (not entirely fetched)
				# :param friend: :class:`aiotfm.friend.Friend` friend after this update
				self.dispatch('friend_disconnected', friend)

			elif TC == 34: # Friend list loaded
				self.friends = FriendList(self, packet)

				# :desc: Called when the friend list is loaded.
				# :param friends: :class:`aiotfm.friend.FriendList` the friend list
				self.dispatch('friends_loaded', self.friends)

			elif TC == 35 or TC == 36: # Friend update / addition
				if self.friends is None:
					return True

				new = Friend(self.friends, packet)
				old = self.friends.get_friend(new.name)

				if old is not None:
					if old.isSoulmate: # Not sent by the server, checked locally.
						self.friends.soulmate = new
						new.isSoulmate = True

					self.friends.friends.remove(old)
				self.friends.friends.append(new)

				if old is None:
					# :desc: Called when a friend is added
					# :param friend: :class:`aiotfm.friend.Friend` the friend
					self.dispatch('new_friend', new)

				else:
					# :desc: Called when a friend is updated
					# :param before: :class:`aiotfm.friend.Friend` friend before this update
					# :param after: :class:`aiotfm.friend.Friend` friend after this update
					self.dispatch('friend_update', old, new)

			elif TC == 37: # Remove friend
				if self.friends is None:
					return True

				friend = self.friends.get_friend(packet.read32())
				if friend is not None:
					if friend == self.friends.soulmate:
						self.friends.soulmate = None

					self.friends.friends.remove(friend)

					# :desc: Called when a friend is removed
					# :param friend: :class:`aiotfm.friend.Friend` the friend
					self.dispatch('friend_remove', friend)

			elif TC == 55: # Channel join result
				sequenceId = packet.read32()
				result = packet.read8()

				# :desc: Called when the client receives the result of joining a channel.
				# :param sequenceId: :class:`int` identifier returned by :meth:`Client.sendCP`.
				# :param result: :class:`int` result code.
				self.dispatch('channel_joined_result', sequenceId, result)

			elif TC == 57: # Channel leave result
				sequenceId = packet.read32()
				result = packet.read8()

				# :desc: Called when the client receives the result of leaving a channel.
				# :param sequenceId: :class:`int` identifier returned by :meth:`Client.sendCP`.
				# :param result: :class:`int` result code.
				self.dispatch('channel_left_result', sequenceId, result)

			elif TC == 59: # Channel /who result
				idSequence = packet.read32()
				result = packet.read8()
				players = [Player(packet.readUTF()) for _ in range(packet.read16())]

				# :desc: Called when the client receives the result of the /who command in a channel.
				# :param idSequence: :class:`int` the reference to the packet that performed the request.
				# :param players: List[:class:`aiotfm.Player`] the list of players inside the channel.
				self.dispatch('channel_who', idSequence, players)

			elif TC == 62: # Joined a channel
				name = packet.readUTF()

				if name in self._channels:
					channel = [c for c in self._channels if c == name][0]
				else:
					channel = Channel(name, self)
					self._channels.append(channel)

				# :desc: Called when the client joined a channel.
				# :param channel: :class:`aiotfm.message.Channel` the channel.
				self.dispatch('channel_joined', channel)

			elif TC == 63: # Quit a channel
				name = packet.readUTF()
				if name in self._channels:
					self._channels.remove(name)

				# :desc: Called when the client leaves a channel.
				# :param name: :class:`str` the channel's name.
				self.dispatch('channel_closed', name)

			elif TC == 64: # Channel message
				username, community = packet.readUTF(), packet.read32()
				channel_name, message = packet.readUTF(), packet.readUTF()
				channel = self.get_channel(channel_name)
				author = self.room.get_player(username=username)

				if author is None:
					author = Player(username)

				if channel is None:
					channel = Channel(channel_name, self)
					self._channels.append(channel)

				channel_message = ChannelMessage(author, community, message, channel)

				# :desc: Called when the client receives a message from a channel.
				# :param message: :class:`aiotfm.message.ChannelMessage` the message.
				self.dispatch('channel_message', channel_message)

			elif TC == 65: # Tribe message
				author, message = packet.readUTF(), packet.readUTF()

				# :desc: Called when the client receives a message from the tribe.
				# :param author: :class:`str` the message's author.
				# :param message: :class:`str` the message's content.
				self.dispatch('tribe_message', author, message)

			elif TC == 66: # Whisper
				author = Player(packet.readUTF())
				commu = packet.read32()
				receiver = Player(packet.readUTF())
				message = packet.readUTF()

				author = self.room.get_player(name=author, default=author)
				receiver = self.room.get_player(name=receiver, default=receiver)

				# :desc: Called when the client receives a whisper.
				# :param message: :class:`aiotfm.message.Whisper` the message.
				self.dispatch('whisper', Whisper(author, commu, receiver, message, self))

			elif TC == 88: # tribe member connected
				# :desc: Called when a tribe member connected.
				# :param name: :class:`str` the member's name.
				self.dispatch('member_connected', packet.readUTF())

			elif TC == 90: # tribe member disconnected
				# :desc: Called when a tribe member disconnected.
				# :param name: :class:`str` the member's name.
				self.dispatch('member_disconnected', packet.readUTF())

			else:
				if self.LOG_UNHANDLED_PACKETS:
					print(CCC, TC, bytes(packet.buffer)[4:])
				return False

		elif CCC == (144, 1): # Set player list
			before = self.room.players
			self.room.players = {}

			for _ in range(packet.read16()):
				player = Player.from_packet(packet)
				self.room.players[player.pid] = player

			# :desc: Called when the client receives an update of all player in the room.
			# :param before: Dict[:class:`aiotfm.Player`] the list of player before the update.
			# :param players: Dict[:class:`aiotfm.Player`] the list of player updated.
			self.dispatch('bulk_player_update', before, self.room.players)

		elif CCC == (144, 2): # Add a player
			after = Player.from_packet(packet)
			before = self.room.players.pop(after.pid, None)

			self.room.players[after.pid] = after
			if before is None:
				# :desc: Called when a player joined the room.
				# :param player: :class:`aiotfm.Player` the player.
				self.dispatch('player_join', after)
			else:
				# :desc: Called when a player's data on the room has been updated.
				# :param before: :class:`aiotfm.Player` the player before the update.
				# :param player: :class:`aiotfm.Player` the player updated.
				self.dispatch('player_update', before, after)

		else:
			if self.LOG_UNHANDLED_PACKETS:
				print(CCC, bytes(packet.buffer)[2:])
			return False

		return True

	async def handle_old_packet(self, connection: Connection, oldCCC: tuple, data: list) -> bool:
		"""|coro|
		Handles the known packets from the old protocol and dispatches events.
		Subclasses should handle only the unhandled packets from this method.

		Example: ::
			class Bot(aiotfm.Client):
				async def handle_old_packet(self, conn, oldCCC, data):
					handled = await super().handle_old_packet(conn, data.copy())

					if not handled:
						# Handle here the unhandled packets.
						pass

		:param connection: :class:`aiotfm.Connection` the connection that received
			the packet.
		:param oldCCC: :class:`tuple` the packet identifiers on the old protocol.
		:param data: :class:`list` the packet data.
		:return: True if the packet got handled, False otherwise.
		"""
		if oldCCC == (8, 5): # Player died
			player = self.room.get_player(pid=data[0])
			if player is not None:
				player.score = int(data[2])

				# :desc: Called when a player dies.
				# :param player: :class:`aiotfm.Player` the player.
				self.dispatch('player_died', player)

		elif oldCCC == (8, 7): # Remove a player
			player = self.room.players.pop(int(data[0]), None)

			if player is not None:
				# :desc: Called when a player leaves the room.
				# :param player: :class:`aiotfm.Player` the player.
				self.dispatch('player_remove', player)

		else:
			if self.LOG_UNHANDLED_PACKETS:
				print("[OLD]", oldCCC, data)
			return False

		return True

	def get_channel(self, name: str) -> Optional[Channel]:
		"""Returns a channel from it's name or None if not found.
		:param name: :class:`str` the name of the channel.
		:return: :class:`aiotfm.message.ChannelMessage` or None
		"""
		if name is None:
			return None

		for channel in self._channels:
			if channel.name == name:
				return channel

	def get_trade(self, player: Union[str, Player]) -> Optional[Trade]:
		"""Returns the pending/current trade with a player.
		:param player: :class:`aiotfm.Player` or :class:`str` the player.
		:return: :class:`aiotfm.inventory.Trade` the trade with the player.
		"""
		if not isinstance(player, (str, Player)):
			raise TypeError(f"Expected Player or str types got {type(player)}")

		if isinstance(player, Player):
			return self.trades.get(player.pid)

		player = player.lower()
		for trade in self.trades.values():
			if trade.trader.lower() == player:
				return trade

	def event(self, coro: Callable) -> Callable:
		"""A decorator that registers an event.

		More about events [here](Events.md).
		"""
		name = coro.__name__
		if not name.startswith('on_'):
			raise InvalidEvent("'{}' isn't a correct event naming.".format(name))
		if not asyncio.iscoroutinefunction(coro):
			message = "Couldn't register a non-coroutine function for the event {}.".format(name)
			raise InvalidEvent(message)

		setattr(self, name, coro)
		return coro

	def wait_for(
		self,
		event: str,
		condition: Optional[Callable] = None,
		timeout: Optional[float] = None,
		stopPropagation: bool = False
	) -> asyncio.Future:
		"""Wait for an event.

		Example: ::
			@client.event
			async def on_room_message(author, message):
				if message == 'id':
					await client.sendCommand(f'profile {author.username}')
					profile = await client.wait_for('on_profile', lambda p: p.username == author)
					await client.sendRoomMessage('Your id: {}'.format(profile.id))

		:param event: :class:`str` the event name.
		:param condition: Optionnal[`function`] A predicate to check what to wait for.
			The arguments must meet the parameters of the event being waited for.
		:param timeout: Optionnal[:class:`float`] the number of seconds before
			throwing asyncio.TimeoutError
		:return: [`asyncio.Future`](https://docs.python.org/3/library/asyncio-future.html#asyncio.Future)
			a future that you must await.
		"""
		event = event.lower()
		future = self.loop.create_future()

		if condition is None:
			def everything(*a):
				return True
			condition = everything

		if event not in self._waiters:
			self._waiters[event] = []

		self._waiters[event].append((condition, future, stopPropagation))

		return asyncio.wait_for(future, timeout)

	async def _run_event(self, coro: Callable, event_name: str, *args, **kwargs):
		"""|coro|
		Runs an event and handle the error if any.

		:param coro: a coroutine function.
		:param event_name: :class:`str` the event's name.
		:param args: arguments to pass to the coro.
		:param kwargs: keyword arguments to pass to the coro.

		:return: :class:`bool` whether the event ran successfully or not
		"""
		try:
			await coro(*args, **kwargs)
			return True
		# except asyncio.CancelledError:
		# 	raise
		except Exception as e:
			if hasattr(self, 'on_error'):
				try:
					await self.on_error(event_name, e, *args, **kwargs)
				# except asyncio.CancelledError:
				# 	raise
				except Exception:
					if self.auto_restart:
						await self.restart(5)
					else:
						self.close()

		return False

	def dispatch(self, event: str, *args, **kwargs):
		"""Dispatches events

		:param event: :class:`str` event's name. (without 'on_')
		:param args: arguments to pass to the coro.
		:param kwargs: keyword arguments to pass to the coro.

		:return: [`Task`](https://docs.python.org/3/library/asyncio-task.html#asyncio.Task)
			the _run_event wrapper task
		"""
		method = 'on_' + event

		if method in self._waiters:
			to_remove = []
			waiters = self._waiters[method]
			for i, (cond, fut, stop) in enumerate(waiters):
				if fut.cancelled():
					to_remove.append(i)
					continue

				try:
					result = bool(cond(*args))
				except Exception as e:
					fut.set_exception(e)
				else:
					if result:
						fut.set_result(args[0] if len(args) == 1 else args if len(args) > 0 else None)
						if stop:
							del waiters[i]
							return None
						to_remove.append(i)

			if len(to_remove) == len(waiters):
				del self._waiters[method]
			else:
				for i in to_remove[::-1]:
					del waiters[i]

		coro = getattr(self, method, None)
		if coro is not None:
			dispatch = self._run_event(coro, method, *args, **kwargs)
			return asyncio.ensure_future(dispatch, loop=self.loop)

	async def on_error(self, event: str, err: Exception, *a, **kw):
		"""Default on_error event handler. Prints the traceback of the error."""
		logger.error('An error occurred while dispatching the event "%s":', event, exc_info=-3)

	async def on_connection_error(self, conn: Connection, error: Exception):
		"""Default on_connection_error event handler. Prints the error."""
		logger.error('The %s connection has been closed.', conn.name, exc_info=error)

	async def on_login_result(self, code: int, *args):
		"""Default on_login_result handler. Raise an error and closes the connection."""
		self.loop.call_later(3, self.close)
		if code == 1:
			raise AlreadyConnected()
		if code == 2:
			raise IncorrectPassword()
		raise LoginError(code)

	async def _connect(self):
		"""|coro|
		Creates a connection with the main server.
		"""
		if self._close_event is None:
			raise AiotfmException(f'{self._connect.__name__} should not be called directly. Use start() instead.')

		for port in random.sample(self.keys.server_ports, 4):
			try:
				await self.main.connect(self.keys.server_ip, port)
			except Exception:
				logger.debug(f'Unable to connect to the server "{self.keys.server_ip}:{port}".')
			else:
				break
		else:
			raise ServerUnreachable('Unable to connect to the server.')

		while not self.main.open:
			await asyncio.sleep(0)

	async def sendHandshake(self):
		"""|coro|
		Sends the handshake packet so the server recognizes this socket as a player.
		"""
		packet = Packet.new(28, 1).write16(self.keys.version)
		if not self.bot_role:
			packet.writeString('en').writeString(self.keys.connection)

		packet.writeString('Desktop').writeString('-').write32(0x1fbd).writeString('')
		packet.writeString('74696720697320676f6e6e61206b696c6c206d7920626f742e20736f20736164')
		packet.writeString(
			"A=t&SA=t&SV=t&EV=t&MP3=t&AE=t&VE=t&ACC=t&PR=t&SP=f&SB=f&DEB=f&V=LNX 32,0,0,182&M=Adobe"
			" Linux&R=1920x1080&COL=color&AR=1.0&OS=Linux&ARCH=x86&L=en&IME=t&PR32=t&PR64=t&LS=en-U"
			"S&PT=Desktop&AVD=f&LFD=f&WD=f&TLS=t&ML=5.1&DP=72")
		packet.write32(0).write32(0x6257).writeString('')

		await self.main.send(packet)

	async def start(
		self,
		api_tfmid: Optional[int] = None,
		api_token: Optional[str] = None,
		keys: Optional[Keys] = None,
		**kwargs
	):
		"""|coro|
		Starts the client.

		:param api_tfmid: Optional[:class:`int`] your Transformice id.
		:param api_token: Optional[:class:`str`] your token to access the API.
		"""
		if self.bot_role:
			self.keys = await get_ip()
		else:
			if self.auto_restart and api_tfmid is None or api_token is None:
				warnings.warn("The api token were not provided. The Client won't be able to restart.")
				self.auto_restart = False

			self.keys = keys
			if keys is None:
				self.keys = await get_keys(api_tfmid, api_token)

		if 'username' in kwargs and 'password' in kwargs:
			# Monkey patch the on_login_ready event
			if hasattr(self, 'on_login_ready'):
				event = getattr(self, 'on_login_ready')
				self.on_login_ready = lambda *a: asyncio.gather(self.login(**kwargs), event(*a))
			else:
				self.on_login_ready = lambda *a: self.login(**kwargs)

		retries = 0
		on_started = None
		keep_alive = Packet.new(26, 26)
		while True:
			self._close_event = asyncio.Future()
			try:
				logger.info('Connecting to the game.')
				await self._connect()
				await self.sendHandshake()
				await self.locale.load()
				retries = 0 # Connection successful
				self._restarting = False
			except Exception as e:
				logger.error('Connection to the server failed.', exc_info=e)
				if on_started is not None:
					on_started.set_exception(e)
				elif retries > self._max_retries:
					raise e
				else:
					retries += 1
					backoff = self._backoff(retries)
					logger.info('Attempt %d failed. Reconnecting in %.2fs', retries, backoff)
					await asyncio.sleep(backoff)
					continue
			else:
				if on_started is not None:
					on_started.set_result(None)

			while not self._close_event.done():
				# Keep the connection(s) alive
				await asyncio.gather(*[c.send(keep_alive) for c in (self.main, self.bulle) if c])
				await asyncio.wait((self._close_event,), timeout=15)

			reason, delay, on_started = self._close_event.result()
			self._close_event = asyncio.Future()

			logger.debug('[Close Event] Reason: %s, Delay: %d, Callback: %s', reason, delay, on_started)
			logger.debug('Will restart: %s', reason != 'stop' and self.auto_restart)

			# clean up
			for conn in (self.main, self.bulle):
				if conn is not None:
					conn.close()

			if reason == 'stop' or not self.auto_restart:
				break

			await asyncio.sleep(delay)

			# If we don't recreate the connection, we won't be able to connect.
			self.main = Connection('main', self, self.loop)
			self.bulle = None

			# Fetch some fresh keys
			if not self.bot_role and (reason != 'restart' or self.keys is None):
				for i in range(self._max_retries):
					try:
						self.keys = await get_keys(api_tfmid, api_token)
						break
					except MaintenanceError:
						if i == 0:
							logger.info('The game is under maintenance.')

						await asyncio.sleep(30)
				else:
					raise MaintenanceError('The game is under heavy maintenance.')

	async def restart_soon(self, delay: float = 5.0, **kwargs):
		"""|coro|
		Restarts the client in several seconds.

		:param delay: :class:`float` the delay before restarting. Default is 5 seconds.
		:param args: arguments to pass to the :meth:`Client.restart` method.
		:param kwargs: keyword arguments to pass to the :meth:`Client.restart` method."""
		warnings.warn('`Client.restart_soon` is deprecated, use `Client.restart` instead.', DeprecationWarning)
		await self.restart(delay, **kwargs)

	async def restart(self, delay: float = 0, keys: Optional[Keys] = None):
		"""|coro|
		Restarts the client.

		:param delay: the delay before restarting. By default, there is no delay.
		:param keys:
		"""
		if not self.auto_restart or self._close_event is None:
			raise AiotfmException(
				'Unable to restart the Client. Either `auto_restart` is set to '
				'False or you have not started the Client using `Client.start`.'
			)

		if self._restarting or self._close_event.done():
			return

		self.keys = keys
		self._restarting = True
		# :desc: Notify when the client restarts.
		self.dispatch("restart")

		restarted = asyncio.Future()
		self._close_event.set_result(('restart', delay, restarted))
		await restarted

	async def login(self, username: str, password: str, encrypted: bool = True, room: str = '*aiotfm'):
		"""|coro|
		Log in the game.

		:param username: :class:`str` the client username.
		:param password: :class:`str` the client password.
		:param encrypted: Optional[:class:`bool`] whether the password is already encrypted or not.
		:param room: Optional[:class:`str`] the room where the client will be logged in.
		"""
		if self._logged:
			raise AiotfmException('You cannot log in twice.')

		self._logged = True
		if not encrypted:
			password = shakikoo(password)

		packet = Packet.new(26, 8).writeString(username).writeString(password)
		packet.writeString("app:/TransformiceAIR.swf/[[DYNAMIC]]/2/[[DYNAMIC]]/4").writeString(room)

		if self.bot_role:
			packet.write16(0).write8(0).writeString('')
		else:
			packet.write32(self.authkey ^ self.keys.auth)
			packet.write8(0).writeString('')
			packet.cipher(self.keys.identification)

		await self.main.send(Packet.new(176, 1).writeUTF(self.community.name))
		await self.main.send(packet.write8(0))

	def run(self, api_tfmid: int, api_token: str, username: str, password: str, **kwargs):
		"""A blocking call that does the event loop initialization for you.

		Equivalent to: ::
			@bot.event
			async def on_login_ready(*a):
				await bot.login(username, password)

			loop = asyncio.get_event_loop()
			loop.create_task(bot.start(api_id, api_token))
			loop.run_forever()
		"""
		try:
			self.loop.run_until_complete(self.start(api_tfmid, api_token, username=username, password=password, **kwargs))
		finally:
			self.loop.run_until_complete(self.loop.shutdown_asyncgens())
			self.loop.close()

	def close(self):
		"""Closes the sockets."""
		if self._closed:
			return

		self._closed = True
		self._close_event.set_result(('stop', 0, None))

	async def sendCP(self, code: int, data: Union[Packet, ByteString] = b'') -> int:
		"""|coro|
		Send a packet to the community platform.

		:param code: :class:`int` the community platform code.
		:param data: :class:`aiotfm.Packet` or :class:`bytes` the data.
		:return: :class:`int` returns the sequence id.
		"""
		self._sequenceId = sid = (self._sequenceId + 1) % 0xFFFFFFFF

		packet = Packet.new(60, 3).write16(code)
		packet.write32(self._sequenceId).writeBytes(data)
		await self.main.send(packet, cipher=True)

		return sid

	async def sendRoomMessage(self, message: str):
		"""|coro|
		Send a message to the room.

		:param message: :class:`str` the content of the message.
		"""
		packet = Packet.new(6, 6).writeString(message)

		await self.bulle.send(packet, cipher=True)

	async def sendTribeMessage(self, message: str):
		"""|coro|
		Send a message to the tribe.

		:param message: :class:`str` the content of the message.
		"""
		await self.sendCP(50, Packet().writeString(message))

	async def sendChannelMessage(self, channel: Union[Channel, str], message: str):
		"""|coro|
		Send a message to a public channel.

		:param channel: :class:`str` the channel's name.
		:param message: :class:`str` the content of the message.
		"""
		if isinstance(channel, Channel):
			channel = channel.name

		return await self.sendCP(48, Packet().writeString(channel).writeString(message))

	async def whisper(self, username: Union[Player, str], message: AnyStr, overflow: bool = False):
		"""|coro|
		Whisper to a player.

		:param username: :class:`str` the player to whisper.
		:param message: :class:`str` the content of the whisper.
		:param overflow: :class:`bool` will send the complete message if True, splitted
			in several messages.
		"""
		if isinstance(username, Player):
			username = username.username

		async def send(msg):
			await self.sendCP(52, Packet().writeString(username).writeString(msg))

		if isinstance(message, str):
			message = message.encode()
		message = message.replace(b'<', b'&lt;').replace(b'>', b'&gt;')

		await send(message[:255])
		for i in range(255, len(message), 255):
			await asyncio.sleep(1)
			await self.whisper(username, message[i:i + 255])

	async def getTribe(self, disconnected: bool = True) -> Optional[Tribe]:
		"""|coro|
		Gets the client's :class:`aiotfm.Tribe` and return it

		:param disconnected: :class:`bool` if True retrieves also the disconnected members.
		:return: :class:`aiotfm.Tribe` or ``None``.
		"""
		sid = await self.sendCP(108, Packet().writeBool(disconnected))

		def is_tribe(tc, packet):
			return (tc == 109 and packet.read32() == sid) or tc == 130

		tc, packet = await self.wait_for('on_raw_cp', is_tribe, timeout=5)
		if tc == 109:
			result = packet.read8()
			if result == 1:
				tc, packet = await self.wait_for('on_raw_cp', lambda tc, p: tc == 130, timeout=5)
			elif result == 17:
				return None
			else:
				raise CommunityPlatformError(118, result)
		return Tribe(packet)

	async def getRoomList(self, gamemode: Union[GameMode, int] = 0, timeout: float = 3) -> Optional[RoomList]:
		"""|coro|
		Get the room list

		:param gamemode: Optional[:class:`aiotfm.enums.GameMode`] the room's gamemode.
		:param timeout: Optional[:class:`int`] timeout in seconds. Defaults to 3 seconds.
		:return: :class:`aiotfm.room.RoomList` the room list for the given gamemode or None
		"""
		await self.main.send(Packet.new(26, 35).write8(int(gamemode)))

		def predicate(roomlist):
			return gamemode == 0 or roomlist.gamemode == gamemode

		try:
			return await self.wait_for('on_room_list', predicate, timeout=timeout)
		except asyncio.TimeoutError:
			return None

	async def playEmote(self, emote: int, flag: str = 'be'):
		"""|coro|
		Play an emote.

		:param emote: :class:`int` the emote's id.
		:param flag: Optional[:class:`str`] the flag for the emote id 10. Defaults to 'be'.
		"""
		packet = Packet.new(8, 1).write8(emote).write32(0)
		if emote == 10:
			packet.writeString(flag)

		await self.bulle.send(packet)

	async def sendSmiley(self, smiley: int):
		"""|coro|
		Makes the client showing a smiley above it's head.

		:param smiley: :class:`int` the smiley's id. (from 0 to 9)
		"""
		if smiley < 0 or smiley > 9:
			raise AiotfmException('Invalid smiley id')

		packet = Packet.new(8, 5).write8(smiley)

		await self.bulle.send(packet)

	async def loadLua(self, lua_code: AnyStr):
		"""|coro|
		Load a lua code in the room.

		:param lua_code: :class:`str` or :class:`bytes` the lua code to send.
		"""
		if isinstance(lua_code, str):
			lua_code = lua_code.encode()

		await self.bulle.send(Packet.new(29, 1).write24(len(lua_code)).writeBytes(lua_code))

	async def sendCommand(self, command: str):
		"""|coro|
		Send a command to the game.

		:param command: :class:`str` the command to send.
		"""
		await self.main.send(Packet.new(6, 26).writeString(command[:255]), cipher=True)

	async def enterTribe(self):
		"""|coro|
		Enter the tribe house
		"""
		await self.main.send(Packet.new(16, 1))

	async def enterTribeHouse(self):
		"""|coro|
		Alias for :meth:`Client.enterTribe`
		"""
		await self.enterTribe()

	async def enterInvTribeHouse(self, author: str):
		"""|coro|
		Join the tribe house of another player after receiving an /inv.

		:param author: :class:`str` the author's username who sent the invitation.
		"""
		await self.main.send(Packet.new(16, 2).writeString(author))

	async def recruit(self, username: Union[Player, str]):
		"""|coro|
		Send a recruit request to a player.

		:param username: :class:`str` the player's username you want to recruit.
		"""
		if isinstance(username, Player):
			username = username.username
		await self.sendCP(78, Packet().writeString(username))

	async def joinRoom(
		self,
		room_name: str,
		password: Optional[str] = None,
		community: Optional[int] = None,
		auto: bool = False
	):
		"""|coro|
		Join a room.
		The event 'on_joined_room' is dispatched when the client has successfully joined the room.

		:param password: :class:`str` if given the client will ignore `community` and `auto` parameters
			and will connect to the room with the given password.
		:param room_name: :class:`str` the room's name.
		:param community: Optional[:class:`int`] the room's community.
		:param auto: Optional[:class:`bool`] joins a random room (I think).
		"""
		if password is not None:
			packet = Packet.new(5, 39).writeString(password).writeString(room_name)
		else:
			packet = Packet.new(5, 38).writeString(Community(community or self.community).name)
			packet.writeString(room_name).writeBool(auto)

		await self.main.send(packet)

	async def joinChannel(self, name: str, permanent: bool = True):
		"""|coro|
		Join a #channel.
		The event 'on_channel_joined' is dispatched when the client has successfully joined
		a channel.

		:param name: :class:`str` the channel's name
		:param permanent: Optional[:class:`bool`] if True (default) the server will automatically
		reconnect the user to this channel when logged in.
		"""
		await self.sendCP(54, Packet().writeString(name).writeBool(permanent))

	async def leaveChannel(self, channel: Union[Channel, str]):
		"""|coro|
		Leaves a #channel.

		:param channel: :class:`aiotfm.message.Channel` channel to leave.
		"""
		if isinstance(channel, Channel):
			name = channel.name
		else:
			name = channel

		await self.sendCP(56, Packet().writeString(name))

	async def requestShopList(self):
		"""|coro|
		Send a request to the server to get the shop list."""
		await self.main.send(Packet.new(8, 20))

	async def startTrade(self, player: Union[Player, str]) -> Trade:
		"""|coro|
		Starts a trade with the given player.

		:param player: :class:`aiotfm.Player` the player to trade with.
		:return: :class:`aiotfm.inventory.Trade` the resulting trade"""
		if isinstance(player, Player) and player.pid == -1:
			player = player.username

		if isinstance(player, str):
			player = self.room.get_player(username=player)
			if player is None:
				raise AiotfmException("The player must be in your room to start a trade.")

		trade = Trade(self, player)

		self.trades[player.pid] = trade
		await trade.accept()
		return trade

	async def requestInventory(self):
		"""|coro|
		Send a request to the server to get the bot's inventory."""
		await self.main.send(Packet.new(31, 1))
